@@ -23,12 +23,14 @@ import math
 import mimetypes
 import os
 import re
+import threading
 import time
 from collections import namedtuple
 from configparser import ConfigParser
 from hashlib import sha256, md5
+from queue import Queue
 from signal import signal, SIGINT, SIGTERM, SIGABRT
-from threading import Event
+from threading import Event, Thread
 
 from pyrogram.api import functions, types
 from pyrogram.api.core import Object
@@ -37,17 +39,17 @@ from pyrogram.api.errors import (
     PhoneNumberUnoccupied, PhoneCodeInvalid, PhoneCodeHashEmpty,
     PhoneCodeExpired, PhoneCodeEmpty, SessionPasswordNeeded,
     PasswordHashInvalid, FloodWait, PeerIdInvalid, FilePartMissing,
-    ChatAdminRequired, FirstnameInvalid
+    ChatAdminRequired, FirstnameInvalid, PhoneNumberBanned
 )
 from pyrogram.api.types import (
     User, Chat, Channel,
-    PeerUser, PeerChat, PeerChannel,
-    Dialog, Message,
+    PeerUser, PeerChannel,
     InputPeerEmpty, InputPeerSelf,
     InputPeerUser, InputPeerChat, InputPeerChannel
 )
 from pyrogram.crypto import CTR
 from pyrogram.session import Auth, Session
+from .input_media import InputMedia
 from .style import Markdown, HTML
 
 log = logging.getLogger(__name__)
@@ -95,11 +97,12 @@ class Client:
             be an empty string: ""
 
         workers (:obj:`int`, optional):
-            Thread pool size for handling incoming messages (updates). Defaults to 4.
+            Thread pool size for handling incoming updates. Defaults to 4.
     """
 
     INVITE_LINK_RE = re.compile(r"^(?:https?://)?t\.me/joinchat/(.+)$")
     DIALOGS_AT_ONCE = 100
+    UPDATES_WORKERS = 2
 
     def __init__(self,
                  session_name: str,
@@ -130,6 +133,8 @@ class Client:
         self.peers_by_id = {}
         self.peers_by_username = {}
 
+        self.channels_pts = {}
+
         self.markdown = Markdown(self.peers_by_id)
         self.html = HTML(self.peers_by_id)
 
@@ -137,9 +142,12 @@ class Client:
         self.proxy = None
         self.session = None
 
-        self.update_handler = None
         self.is_idle = Event()
 
+        self.updates_queue = Queue()
+        self.update_queue = Queue()
+        self.update_handler = None
+
     def start(self):
         """Use this method to start the Client after creating it.
         Requires no parameters.
@@ -156,7 +164,7 @@ class Client:
             self.proxy,
             self.auth_key,
             self.config.api_id,
-            workers=self.workers
+            client=self
         )
 
         terms = self.session.start()
@@ -170,7 +178,12 @@ class Client:
 
         self.rnd_id = self.session.msg_id
         self.get_dialogs()
-        self.session.update_handler = self.update_handler
+
+        for i in range(self.UPDATES_WORKERS):
+            Thread(target=self.updates_worker, name="UpdatesWorker#{}".format(i + 1)).start()
+
+        for i in range(self.workers):
+            Thread(target=self.update_worker, name="UpdateWorker#{}".format(i + 1)).start()
 
         mimetypes.init()
 
@@ -180,6 +193,163 @@ class Client:
         """
         self.session.stop()
 
+        for _ in range(self.UPDATES_WORKERS):
+            self.updates_queue.put(None)
+
+        for _ in range(self.workers):
+            self.update_queue.put(None)
+
+    def fetch_peers(self, entities: list):
+        for entity in entities:
+            if isinstance(entity, User):
+                user_id = entity.id
+
+                if user_id in self.peers_by_id:
+                    continue
+
+                access_hash = entity.access_hash
+
+                if access_hash is None:
+                    continue
+
+                username = entity.username
+
+                input_peer = InputPeerUser(
+                    user_id=user_id,
+                    access_hash=access_hash
+                )
+
+                self.peers_by_id[user_id] = input_peer
+
+                if username is not None:
+                    self.peers_by_username[username] = input_peer
+
+            if isinstance(entity, Chat):
+                chat_id = entity.id
+
+                if chat_id in self.peers_by_id:
+                    continue
+
+                input_peer = InputPeerChat(
+                    chat_id=chat_id
+                )
+
+                self.peers_by_id[chat_id] = input_peer
+
+            if isinstance(entity, Channel):
+                channel_id = entity.id
+                peer_id = int("-100" + str(channel_id))
+
+                if peer_id in self.peers_by_id:
+                    continue
+
+                access_hash = entity.access_hash
+
+                if access_hash is None:
+                    continue
+
+                username = entity.username
+
+                input_peer = InputPeerChannel(
+                    channel_id=channel_id,
+                    access_hash=access_hash
+                )
+
+                self.peers_by_id[peer_id] = input_peer
+
+                if username is not None:
+                    self.peers_by_username[username] = input_peer
+
+    def updates_worker(self):
+        name = threading.current_thread().name
+        log.debug("{} started".format(name))
+
+        while True:
+            updates = self.updates_queue.get()
+
+            if updates is None:
+                break
+
+            try:
+                if isinstance(updates, (types.Update, types.UpdatesCombined)):
+                    self.fetch_peers(updates.users)
+                    self.fetch_peers(updates.chats)
+
+                    for update in updates.updates:
+                        channel_id = getattr(
+                            getattr(
+                                getattr(
+                                    update, "message", None
+                                ), "to_id", None
+                            ), "channel_id", None
+                        ) or getattr(update, "channel_id", None)
+
+                        pts = getattr(update, "pts", None)
+
+                        if channel_id and pts:
+                            if channel_id not in self.channels_pts:
+                                self.channels_pts[channel_id] = []
+
+                            if pts in self.channels_pts[channel_id]:
+                                continue
+
+                            self.channels_pts[channel_id].append(pts)
+
+                            if len(self.channels_pts[channel_id]) > 50:
+                                self.channels_pts[channel_id] = self.channels_pts[channel_id][25:]
+
+                        self.update_queue.put((update, updates.users, updates.chats))
+                elif isinstance(updates, (types.UpdateShortMessage, types.UpdateShortChatMessage)):
+                    diff = self.send(
+                        functions.updates.GetDifference(
+                            pts=updates.pts - updates.pts_count,
+                            date=updates.date,
+                            qts=-1
+                        )
+                    )
+
+                    self.fetch_peers(diff.users)
+                    self.fetch_peers(diff.chats)
+
+                    self.update_queue.put((
+                        types.UpdateNewMessage(
+                            message=diff.new_messages[0],
+                            pts=updates.pts,
+                            pts_count=updates.pts_count
+                        ),
+                        diff.users,
+                        diff.chats
+                    ))
+                elif isinstance(updates, types.UpdateShort):
+                    self.update_queue.put((updates.update, [], []))
+            except Exception as e:
+                log.error(e, exc_info=True)
+
+        log.debug("{} stopped".format(name))
+
+    def update_worker(self):
+        name = threading.current_thread().name
+        log.debug("{} started".format(name))
+
+        while True:
+            update = self.update_queue.get()
+
+            if update is None:
+                break
+
+            try:
+                if self.update_handler:
+                    self.update_handler(
+                        self,
+                        update[0],
+                        {i.id: i for i in update[1]},
+                        {i.id: i for i in update[2]}
+                    )
+            except Exception as e:
+                log.error(e, exc_info=True)
+
+        log.debug("{} stopped".format(name))
+
     def signal_handler(self, *args):
         self.stop()
         self.is_idle.set()
@@ -198,18 +368,49 @@ class Client:
 
         self.is_idle.wait()
 
-    # TODO: Better update handler
     def set_update_handler(self, callback: callable):
         """Use this method to set the update handler.
 
+        You must call this method *before* you *start()* the Client.
+
         Args:
             callback (:obj:`callable`):
-                A callback function that accepts a single argument: the update object.
+                A function that will be called when a new update is received from the server. It takes
+                :obj:`(client, update, users, chats)` as positional arguments (Look at the section below for
+                a detailed description).
+
+        Other Parameters:
+            client (:obj:`pyrogram.Client`):
+                The Client itself, useful when you want to call other API methods inside the update handler.
+
+            update (:obj:`Update`):
+                The received update, which can be one of the many single Updates listed in the *updates*
+                field you see in the :obj:`types.Update <pyrogram.api.types.Update>` type.
+
+            users (:obj:`dict`):
+                Dictionary of all :obj:`types.User <pyrogram.api.types.User>` mentioned in the update.
+                You can access extra info about the user (such as *first_name*, *last_name*, etc...) by using
+                the IDs you find in the *update* argument (e.g.: *users[1768841572]*).
+
+            chats (:obj:`dict`):
+                Dictionary of all :obj:`types.Chat <pyrogram.api.types.Chat>` and
+                :obj:`types.Channel <pyrogram.api.types.Channel>` mentioned in the update.
+                You can access extra info about the chat (such as *title*, *participants_count*, etc...)
+                by using the IDs you find in the *update* argument (e.g.: *chats[1701277281]*).
+
+        Note:
+            The following Empty or Forbidden types may exist inside the *users* and *chats* dictionaries.
+            They mean you have been blocked by the user or banned from the group/channel.
+
+            - :obj:`types.UserEmpty <pyrogram.api.types.UserEmpty>`
+            - :obj:`types.ChatEmpty <pyrogram.api.types.ChatEmpty>`
+            - :obj:`types.ChatForbidden <pyrogram.api.types.ChatForbidden>`
+            - :obj:`types.ChannelForbidden <pyrogram.api.types.ChannelForbidden>`
         """
         self.update_handler = callback
 
     def send(self, data: Object):
-        """Use this method to send :ref:`Raw Function <using-raw-functions>` queries.
+        """Use this method to send Raw Function queries.
 
         This method makes possible to manually call every single Telegram API method in a low-level manner.
         Available functions are listed in the :obj:`pyrogram.api.functions` package and may accept compound
@@ -262,7 +463,7 @@ class Client:
                     self.proxy,
                     self.auth_key,
                     self.config.api_id,
-                    workers=self.workers
+                    client=self
                 )
                 self.session.start()
 
@@ -274,7 +475,7 @@ class Client:
                     )
                 )
                 break
-            except PhoneNumberInvalid as e:
+            except (PhoneNumberInvalid, PhoneNumberBanned) as e:
                 if phone_number_invalid_raises:
                     raise
                 else:
@@ -432,53 +633,17 @@ class Client:
             )
 
     def get_dialogs(self):
-        peers = []
+        def parse_dialogs(d):
+            self.fetch_peers(d.chats)
+            self.fetch_peers(d.users)
 
-        def parse_dialogs(d) -> int:
-            oldest_date = 1 << 32
-
-            for dialog in d.dialogs:  # type: Dialog
-                # Only search for Users, Chats and Channels
-                if not isinstance(dialog.peer, (PeerUser, PeerChat, PeerChannel)):
+            for m in reversed(d.messages):
+                if isinstance(m, types.MessageEmpty):
                     continue
-
-                if isinstance(dialog.peer, PeerUser):
-                    peer_type = "user"
-                    peer_id = dialog.peer.user_id
-                elif isinstance(dialog.peer, PeerChat):
-                    peer_type = "chat"
-                    peer_id = dialog.peer.chat_id
-                elif isinstance(dialog.peer, PeerChannel):
-                    peer_type = "channel"
-                    peer_id = dialog.peer.channel_id
                 else:
-                    continue
-
-                for message in d.messages:  # type: Message
-                    is_this = peer_id == message.from_id or dialog.peer == message.to_id
-
-                    if is_this:
-                        for entity in (d.users if peer_type == "user" else d.chats):  # type: User or Chat or Channel
-                            if entity.id == peer_id:
-                                peers.append(
-                                    dict(
-                                        id=peer_id,
-                                        access_hash=getattr(entity, "access_hash", None),
-                                        type=peer_type,
-                                        first_name=getattr(entity, "first_name", None),
-                                        last_name=getattr(entity, "last_name", None),
-                                        title=getattr(entity, "title", None),
-                                        username=getattr(entity, "username", None),
-                                    )
-                                )
-
-                                if message.date < oldest_date:
-                                    oldest_date = message.date
-
-                                break
-                        break
-
-            return oldest_date
+                    return m.date
+            else:
+                return 0
 
         pinned_dialogs = self.send(functions.messages.GetPinnedDialogs())
         parse_dialogs(pinned_dialogs)
@@ -491,48 +656,23 @@ class Client:
         )
 
         offset_date = parse_dialogs(dialogs)
-        log.info("Dialogs count: {}".format(len(peers)))
+        log.info("Entities count: {}".format(len(self.peers_by_id)))
 
         while len(dialogs.dialogs) == self.DIALOGS_AT_ONCE:
-            dialogs = self.send(
-                functions.messages.GetDialogs(
-                    offset_date, 0, types.InputPeerEmpty(),
-                    self.DIALOGS_AT_ONCE, True
+            try:
+                dialogs = self.send(
+                    functions.messages.GetDialogs(
+                        offset_date, 0, types.InputPeerEmpty(),
+                        self.DIALOGS_AT_ONCE, True
+                    )
                 )
-            )
-
-            offset_date = parse_dialogs(dialogs)
-            log.info("Dialogs count: {}".format(len(peers)))
-
-        for i in peers:
-            peer_id = i["id"]
-            peer_type = i["type"]
-            peer_username = i["username"]
-            peer_access_hash = i["access_hash"]
-
-            if peer_type == "user":
-                input_peer = InputPeerUser(
-                    peer_id,
-                    peer_access_hash
-                )
-            elif peer_type == "chat":
-                input_peer = InputPeerChat(
-                    peer_id
-                )
-            elif peer_type == "channel":
-                input_peer = InputPeerChannel(
-                    peer_id,
-                    peer_access_hash
-                )
-                peer_id = int("-100" + str(peer_id))
-            else:
+            except FloodWait as e:
+                log.info("Get dialogs flood wait: {}".format(e.x))
+                time.sleep(e.x)
                 continue
 
-            self.peers_by_id[peer_id] = input_peer
-
-            if peer_username:
-                peer_username = peer_username.lower()
-                self.peers_by_username[peer_username] = input_peer
+            offset_date = parse_dialogs(dialogs)
+            log.info("Entities count: {}".format(len(self.peers_by_id)))
 
     def resolve_username(self, username: str):
         username = username.lower().strip("@")
@@ -564,24 +704,51 @@ class Client:
         return input_peer
 
     def resolve_peer(self, peer_id: int or str):
-        if peer_id in ("self", "me"):
-            return InputPeerSelf()
-        else:
-            if type(peer_id) is str:
-                peer_id = peer_id.lower().strip("@")
+        """Use this method to get the *InputPeer* of a known *peer_id*.
 
-                try:
-                    return self.peers_by_username[peer_id]
-                except KeyError:
-                    return self.resolve_username(peer_id)
-            else:
-                try:
-                    return self.peers_by_id[peer_id]
-                except KeyError:
-                    try:
-                        return self.peers_by_id[int("-100" + str(peer_id))]
-                    except KeyError:
-                        raise PeerIdInvalid
+        It is intended to be used when working with Raw Functions (i.e: a Telegram API method you wish to use which is
+        not available yet in the Client class as an easy-to-use method).
+
+        Args:
+            peer_id (:obj:`int` | :obj:`str` | :obj:`Peer`):
+                The Peer ID you want to extract the InputPeer from. Can be one of these types: :obj:`int` (direct ID),
+                :obj:`str` (@username), :obj:`PeerUser <pyrogram.api.types.PeerUser>`,
+                :obj:`PeerChat <pyrogram.api.types.PeerChat>`, :obj:`PeerChannel <pyrogram.api.types.PeerChannel>`
+
+        Returns:
+            :obj:`InputPeerUser <pyrogram.api.types.InputPeerUser>` or
+            :obj:`InputPeerChat <pyrogram.api.types.InputPeerChat>` or
+            :obj:`InputPeerChannel <pyrogram.api.types.InputPeerChannel>` depending on the *peer_id*.
+
+        Raises:
+            :class:`pyrogram.Error`
+        """
+        if type(peer_id) is str:
+            if peer_id in ("self", "me"):
+                return InputPeerSelf()
+
+            peer_id = peer_id.lower().strip("@")
+
+            try:
+                return self.peers_by_username[peer_id]
+            except KeyError:
+                return self.resolve_username(peer_id)
+
+        if type(peer_id) is not int:
+            if isinstance(peer_id, types.PeerUser):
+                peer_id = peer_id.user_id
+            elif isinstance(peer_id, types.PeerChat):
+                peer_id = peer_id.chat_id
+            elif isinstance(peer_id, types.PeerChannel):
+                peer_id = int("-100" + str(peer_id.channel_id))
+
+        try:
+            return self.peers_by_id[peer_id]
+        except KeyError:
+            try:
+                return self.peers_by_id[int("-100" + str(peer_id))]
+            except KeyError:
+                raise PeerIdInvalid
 
     def get_me(self):
         """A simple method for testing the user authorization. Requires no parameters.
@@ -604,7 +771,7 @@ class Client:
                      parse_mode: str = "",
                      disable_web_page_preview: bool = None,
                      disable_notification: bool = None,
-                     reply_to_msg_id: int = None):
+                     reply_to_message_id: int = None):
         """Use this method to send text messages.
 
         Args:
@@ -628,7 +795,7 @@ class Client:
                 Sends the message silently.
                 Users will receive a notification with no sound.
 
-            reply_to_msg_id (:obj:`bool`, optional):
+            reply_to_message_id (:obj:`bool`, optional):
                 If the message is a reply, ID of the original message.
 
         Returns:
@@ -644,7 +811,7 @@ class Client:
                 peer=self.resolve_peer(chat_id),
                 no_webpage=disable_web_page_preview or None,
                 silent=disable_notification or None,
-                reply_to_msg_id=reply_to_msg_id,
+                reply_to_msg_id=reply_to_message_id,
                 random_id=self.rnd_id(),
                 **style.parse(text)
             )
@@ -919,6 +1086,7 @@ class Client:
                    duration: int = 0,
                    width: int = 0,
                    height: int = 0,
+                   supports_streaming: bool = None,
                    disable_notification: bool = None,
                    reply_to_message_id: int = None):
         """Use this method to send video files.
@@ -950,6 +1118,9 @@ class Client:
             height (:obj:`int`, optional):
                 Video height.
 
+            supports_streaming (:obj:`bool`, optional):
+                Pass True, if the uploaded video is suitable for streaming.
+
             disable_notification (:obj:`bool`, optional):
                 Sends the message silently.
                 Users will receive a notification with no sound.
@@ -976,10 +1147,12 @@ class Client:
                             file=file,
                             attributes=[
                                 types.DocumentAttributeVideo(
+                                    supports_streaming=supports_streaming,
                                     duration=duration,
                                     w=width,
                                     h=height
-                                )
+                                ),
+                                types.DocumentAttributeFilename(os.path.basename(video))
                             ]
                         ),
                         silent=disable_notification or None,
@@ -1941,3 +2114,101 @@ class Client:
             )
         else:
             return False
+
+    def send_media_group(self,
+                         chat_id: int or str,
+                         media: list,
+                         disable_notification: bool = None,
+                         reply_to_message_id: int = None):
+        """Use this method to send a group of photos or videos as an album.
+        On success, an Update containing the sent Messages is returned.
+
+        Args:
+            chat_id (:obj:`int` | :obj:`str`):
+                Unique identifier for the target chat or username of the target channel/supergroup
+                (in the format @username). For your personal cloud storage (Saved Messages) you can
+                simply use "me" or "self".
+
+            media (:obj:`list`):
+                A list containing either :obj:`pyrogram.InputMedia.Photo` or :obj:`pyrogram.InputMedia.Video` objects
+                describing photos and videos to be sent, must include 2–10 items.
+
+            disable_notification (:obj:`bool`, optional):
+                Sends the message silently.
+                Users will receive a notification with no sound.
+
+            reply_to_message_id (:obj:`int`, optional):
+                If the message is a reply, ID of the original message.
+        """
+        multi_media = []
+
+        for i in media:
+            if isinstance(i, InputMedia.Photo):
+                style = self.html if i.parse_mode.lower() == "html" else self.markdown
+                media = self.save_file(i.media)
+
+                media = self.send(
+                    functions.messages.UploadMedia(
+                        peer=self.resolve_peer(chat_id),
+                        media=types.InputMediaUploadedPhoto(
+                            file=media
+                        )
+                    )
+                )
+
+                single_media = types.InputSingleMedia(
+                    media=types.InputMediaPhoto(
+                        id=types.InputPhoto(
+                            id=media.photo.id,
+                            access_hash=media.photo.access_hash
+                        )
+                    ),
+                    random_id=self.rnd_id(),
+                    **style.parse(i.caption)
+                )
+
+                multi_media.append(single_media)
+            elif isinstance(i, InputMedia.Video):
+                style = self.html if i.parse_mode.lower() == "html" else self.markdown
+                media = self.save_file(i.media)
+
+                media = self.send(
+                    functions.messages.UploadMedia(
+                        peer=self.resolve_peer(chat_id),
+                        media=types.InputMediaUploadedDocument(
+                            file=media,
+                            mime_type=mimetypes.types_map[".mp4"],
+                            attributes=[
+                                types.DocumentAttributeVideo(
+                                    supports_streaming=i.supports_streaming,
+                                    duration=i.duration,
+                                    w=i.width,
+                                    h=i.height
+                                ),
+                                types.DocumentAttributeFilename(os.path.basename(i.media))
+                            ]
+                        )
+                    )
+                )
+
+                single_media = types.InputSingleMedia(
+                    media=types.InputMediaDocument(
+                        id=types.InputDocument(
+                            id=media.document.id,
+                            access_hash=media.document.access_hash
+                        )
+                    ),
+                    random_id=self.rnd_id(),
+                    **style.parse(i.caption)
+                )
+
+                multi_media.append(single_media)
+
+        return self.send(
+            functions.messages.SendMultiMedia(
+                peer=self.resolve_peer(chat_id),
+                multi_media=multi_media,
+                silent=disable_notification or None,
+                reply_to_msg_id=reply_to_message_id
+            )
+        )
